#!/usr/bin/env python
# encoding: utf-8
"""
mixins.py

several mixins that provide additional functionality to the Trainer class

Copyright (c) 2019, I. Manakov

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and
to permit persons to whom the Software is furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in all copies or substantial portions
of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""

import os
from functools import wraps
from time import time, ctime
import warnings

import h5py
import torch as pt
from trainer.utils import to_numpy


class SaveMixin(object):
    """used to automatically collect objects of specified class and save them using pt.save"""

    def save(self, *args,  save_config=None, checkpoint_name=None, directory='./', **kwargs):
        """
        create a checkpoint with all objects of the types specified in the save config
        :param save_config: dict that specifies which attributes to save. Keys should be classes and values conversion methods.
        :param checkpoint_name: name of the checkpoint that will be created
        :param directory: folder where checkpoints will be saved
        :return: filename of the saved checkpoint
        """

        if save_config is None:
            save_config = self.default_save_config().copy()

        # find all objects that need to be saved
        objects = {}

        for key, value in self.__dict__.items():

            # check if value is instance of classes in save_config
            cls = None
            conversion_name = None
            for obj, call in save_config.items():
                if isinstance(value, obj):
                    cls = obj
                    conversion_name = call
                    break

            # apply conversion method specified in save_config for that class
            if cls is not None and conversion_name is not None:

                # find conversion method for present value
                for level in (value, self, object, type):
                    conversion = getattr(level, conversion_name, None)
                    if conversion is not None:
                        break

                # convert the value for storage
                objects[key] = conversion() if level is value else conversion(value)

        # set default save path if it is None
        if checkpoint_name is not None:
            checkpoint = f'checkpoint_{checkpoint_name}.pt'
        else:
            checkpoint = f'checkpoint_{ctime().replace(" ", "-")}.pt'

        # save the objects
        pt.save(objects, os.path.join(directory, checkpoint))
        return checkpoint

    @staticmethod
    def to_numpy(sample):
        """
        convert individual and sequences of Tensors to numpy, while leaving strings, integers and floats unchanged.
        set up as bound method because of the way conversion method lookup works in save
        :param sample: individual or sequence of Tensors, ints, floats or strings
        :return: same as sample but all Tensors are converted to numpy.ndarray
        """
        return to_numpy(sample)

    @staticmethod
    def default_save_config():
        """initialize the default save config that saves pytorch Modules, Tensors and Optimizers
         as well as strings, floats and ints. If apex is installed FP16_Optimizers are also saved"""
        save_config = {pt.nn.Module: 'state_dict',
                       pt.optim.Optimizer: 'state_dict',
                       pt.Tensor: '_to_numpy',
                       int: '__int__',
                       # str: '__str__',
                       float: '__float__'}

        try:
            from apex.fp16_utils import FP16_Optimizer
            save_config[FP16_Optimizer] = 'state_dict'
        except ImportError:
            pass

        return save_config


class MonitorMixin(object):
    """
    wraps public methods to intercept outputs and store them in a hdf5 file
    """

    def __del__(self):
        """make sure that storage is closed on deconstruction"""
        if hasattr(self, 'storage') and hasattr(self.storage, 'close'):
            self.storage.close()

    @staticmethod
    def open_storage(*, filename):
        """
        creates or opens the hdf5 storage and appends it to instance attributes
        :param filename: name of the hdf5 storage
        :return: h5py.File instance
        """

        # create directory if necessary
        if not os.path.isdir(os.path.dirname(filename)):
            os.makedirs(os.path.dirname(filename))

        # open hdf5 file
        if os.path.isfile(filename):
            mode = 'a'
        else:
            mode = 'w'

        storage = h5py.File(filename, mode)#, libver='latest', swmr=True)

        return storage

    def close_storage(self, *args, **kwargs):
        """
        generic method for closing storage
        :param args: not used
        :param kwargs: not used
        :return: None
        """

        storage = getattr(self, 'storage', None)
        if storage is not None:
            storage.close()

    def monitor(self, *, name, method=None):
        """
        wrap a funciton to intercept its results. If only a name is specified, method is retrieved from self and
        will be replaced with wrapped version. If name and method are specified wrapped function will be returned instead.
        :param name: name of the function / method
        :param method: callable to wrap. If this argument is specified, the passed callable will be wrapped and returned.
        Otherwilse, the method will be retrieved from self via getattr and replaced with its wrapped version.
        :return: wrapped callable if method is not None else None
        """
        set_method = False
        if method is None:
            method = getattr(self, name)
            set_method = True
        method = self._wrap(method, name)

        if set_method:
            setattr(self, name, method)
        else:
            return method

    def _wrap(self, func, name):
        """
        intercept return values of func and store them in the hdf5 file before returning them
        :param func: method to wrap
        :return: wrapped method
        """

        @wraps(func)
        def monitored(*args, key=None, **kwargs):

            results = func(*args, **kwargs)

            if key is None:
                key = time()

            # only write to storage if method returns something
            if results is not None:
                self._store(results, name, key)

            return results

        return monitored

    def _store(self, results, group_name, key):
        """
        writes results to specified group in the hdf5 file
        :param results: results as returned by some method
        :param group_name: name of hdf5 group object in the hdf5 file
        """

        storage = getattr(self, 'storage', None)

        # check if storage is initialized
        if storage is None:
            raise ValueError(f'tried to write to group {group_name} under key {key} but storage attribute is None')
        elif isinstance(storage, str):
            storage = self.open_storage(filename=storage)
            setattr(self, 'storage', storage)

        # check if storage is open
        if storage.name is None:
            raise ValueError(f'tried to write to group {group_name} under key {key} but storage was already closed')

        key = f'{key}'
        group = storage.require_group(group_name)

        results = to_numpy(results)

        # case where the result is a namedtuple
        if isinstance(results, tuple) and hasattr(results, '_fields'):

            # obtain fields from namedtuple
            for field in getattr(results, '_fields'):
                field_key = '/'.join([key, field])
                group[field_key] = getattr(results, field)

        # default case
        else:
            group[key] = results

        storage.flush()


class CheckpointMixin(object):

    def load(self, checkpoint, **kwargs):
        """
        load a checkpoint that can contain model and optimizer state
        :param checkpoint: filename of the checkpoint
        :return: None
        """

        checkpoint = pt.load(checkpoint)

        # load components
        for key in checkpoint.keys():

            if not hasattr(self, key) or not checkpoint[key]:
                continue

            print(f'loading {key} checkpoint...', end='')
            if key == 'model' or key == 'optimzer':
                getattr(self, key).load_state_dict(checkpoint[key])
            else:
                setattr(self, key, checkpoint[key])
            print('done')

    @staticmethod
    def list_checkpoints(directory):

        checkpoints = [os.path.join(directory, file) for file in os.listdir(directory) if 'checkpoint' in file]
        checkpoints.sort(key=os.path.getmtime)

        return checkpoints

    def load_latest(self, directory, **kwargs):

        checkpoints = self.list_checkpoints(directory)
        if checkpoints:
            self.load(checkpoints[-1])
        else:
            warnings.warn(f'Checkpoint could not be loaded. No checkpoints found in {directory}.')



from collections import Sequence, namedtuple
from functools import wraps

import numpy as np
import torch as pt
import torch.nn as nn
import torch.nn.init as init


def weight_init(m):
    """
    Usage:
        model = Model()
        model.apply(weight_init)
    """

    if isinstance(m, nn.Conv1d):
        init.normal_(m.weight.data)
        if m.bias is not None:
            init.normal_(m.bias.data)
    elif isinstance(m, nn.Conv2d):
        init.xavier_normal_(m.weight.data)
        if m.bias is not None:
            init.normal_(m.bias.data)
    elif isinstance(m, nn.Conv3d):
        init.xavier_normal_(m.weight.data)
        if m.bias is not None:
            init.normal_(m.bias.data)
    elif isinstance(m, nn.ConvTranspose1d):
        init.normal_(m.weight.data)
        if m.bias is not None:
            init.normal_(m.bias.data)
    elif isinstance(m, nn.ConvTranspose2d):
        init.xavier_normal_(m.weight.data)
        if m.bias is not None:
            init.normal_(m.bias.data)
    elif isinstance(m, nn.ConvTranspose3d):
        init.xavier_normal_(m.weight.data)
        if m.bias is not None:
            init.normal_(m.bias.data)
    elif isinstance(m, nn.BatchNorm1d):
        init.normal_(m.weight.data, mean=1, std=0.02)
        init.constant_(m.bias.data, 0)
    elif isinstance(m, nn.BatchNorm2d):
        init.normal_(m.weight.data, mean=1, std=0.02)
        init.constant_(m.bias.data, 0)
    elif isinstance(m, nn.BatchNorm3d):
        init.normal_(m.weight.data, mean=1, std=0.02)
        init.constant_(m.bias.data, 0)
    elif isinstance(m, nn.Linear):
        init.xavier_normal_(m.weight.data)
        init.normal_(m.bias.data)
    elif isinstance(m, nn.LSTM):
        for param in m.parameters():
            if len(param.shape) >= 2:
                init.orthogonal_(param.data)
            else:
                init.normal_(param.data)
    elif isinstance(m, nn.LSTMCell):
        for param in m.parameters():
            if len(param.shape) >= 2:
                init.orthogonal_(param.data)
            else:
                init.normal_(param.data)
    elif isinstance(m, nn.GRU):
        for param in m.parameters():
            if len(param.shape) >= 2:
                init.orthogonal_(param.data)
            else:
                init.normal_(param.data)
    elif isinstance(m, nn.GRUCell):
        for param in m.parameters():
            if len(param.shape) >= 2:
                init.orthogonal_(param.data)
            else:
                init.normal_(param.data)


def identity(x):
    return x


def to_numpy(sample):
    """
    convert individual and sequences of Tensors to numpy, while leaving strings, integers and floats unchanged
    :param sample: individual or sequence of Tensors, ints, floats or strings
    :return: same as sample but all Tensors are converted to numpy.ndarray
    """
    if isinstance(sample, str):
        return sample
    elif isinstance(sample, pt.Tensor):
        return sample.detach().cpu().numpy()
    elif isinstance(sample, tuple) and hasattr(sample, '_fields'):  # namedtuple
        collection = sample.__class__
        return collection(*[to_numpy(s) for s in sample])
    elif isinstance(sample, Sequence):
        return [to_numpy(s) for s in sample]
    else:
        return sample


def set_seed(seed):
    pt.backends.cudnn.deterministic = True
    pt.backends.cudnn.benchmark = False
    np.random.seed(seed)
    pt.manual_seed(seed)
    pt.cuda.manual_seed_all(seed)


def test_on_sample(*args, sample, forward_pass, **kwargs):
    """
    perform inference on a sample
    :param sample: pt.Tensor on which inference is performed
    :param forward_pass: callable that returns (prediction, loss)
    :return: result of inference as namedtuple('prediction', 'loss')
    """

    with pt.no_grad():
        prediction, loss = to_numpy(forward_pass(sample))

    return namedtuple('test_on_sample', ('prediction', 'loss'))(prediction, loss)


def validate(*args, dataloader, forward_pass, **kwargs):
    """
    evaluate model on validation set
    :param dataloader: Dataloader with validation set
    :param forward_pass: callable that returns (prediction, loss)
    :return: ndarray of losses
    """

    losses = []
    for sample in dataloader:

        with pt.no_grad():
            _, loss = forward_pass(sample)
            losses.append(to_numpy(loss))

    return np.stack(losses)


def show_progress(epoch, step, n_epochs, n_steps, steps_in_epoch, loss):
    """
    print the current progress and loss
    :param epoch: current epoch
    :param step: current step
    :param n_epochs: number of epoch in training
    :param n_steps: number of steps in training
    :param steps_in_epoch: number of steps in one epoch
    :param loss: current loss
    """

    n_steps = n_steps if n_steps is not None else n_epochs * steps_in_epoch - 1
    steps_taken = epoch * steps_in_epoch + step
    progress = round(100 * steps_taken / n_steps, 2)

    print(f'progress: {progress}%, epoch: {epoch}, step: {step}, loss: {loss}')


class IntervalBased(object):
    """
    used for decorating functions.
    the function will then only be executed at set intervals
    """
    def __init__(self, interval):
        """
        :param interval: interval at which the function will be called
        """
        self.interval = interval
        self.counter = 0

    def __call__(self, func):
        """
        wraps function so that it is executed only at set interval
        :param func: function to wrap
        :return: wrapped function
        """
        @wraps(func)
        def wrapped(*args, **kwargs):
            self.counter += 1
            if self.counter % self.interval == 0:
                return func(*args, **kwargs)

        return wrapped



from torch.optim import Optimizer

from trainer.events import AbstractEventHandler
from trainer.mixins import SaveMixin


class EventSave(AbstractEventHandler):
    """
    Adapter for the save method that constructs checkpoint name from epoch and step
    and passes logdir as directory
    """

    def __init__(self, *, func=SaveMixin.save, name=None, interval=None):
        super().__init__(func, name=name, interval=interval)

    def __call__(self, trainer, key=None, loss=None, step=None, epoch=None):
        checkpoint_name = f'epoch-{epoch}' if step is None else f'epoch-{epoch}_step-{step}'
        save_config = SaveMixin.default_save_config().copy()
        save_config.pop(Optimizer)
        self.func(trainer, directory=trainer.logdir, checkpoint_name=checkpoint_name, save_config=save_config)
        return checkpoint_name


class TrainingLoss(AbstractEventHandler):
    """
    Adapter for recording training loss. Simply takes in the loss at each step and returns it.
    If registered as an event handler and monitored the loss will be saved to storage
    """

    def __init__(self, func=lambda x: x, name='training-loss', interval=None):
        super().__init__(func, name=name, interval=interval)

    def __call__(self, trainer, key=None, loss=None, step=None, epoch=None):
        return self.func(loss)




import shutil
import inspect
import importlib
from types import FunctionType, BuiltinFunctionType
import ast
from collections import Sequence


class Imports(object):
    """
    class for handling imported names in a python module
    """

    def __init__(self, imports):
        """
        extract imported names from source code string or another Imports instance
        :param imports: source code of a module of other instance of Imports
        """

        if isinstance(imports, str):
            self.imports, self.aliases = Imports._parse_source_string(imports)
        elif isinstance(imports, Imports):
            self.imports, self.aliases = imports.imports, imports.aliases
        else:
            raise ValueError('input has to be either string or Imports instance')

    @classmethod
    def from_file(cls, file):
        """
        initialize instance from .py config file
        :param file: filename of the config .py file
        :return: Config instance
        """

        spec = importlib.util.spec_from_file_location("config", file)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)

        return cls.from_module(module)

    @classmethod
    def from_module(cls, module):
        """
        initialize Imports instance from module type
        :param module: python module
        :return: Imports instance
        """
        source = inspect.getsource(module)
        return cls(source)

    @staticmethod
    def _parse_source_string(import_string):
        """
        extract imports from a source code string using python's ast
        :param import_string: source code string
        :return: dict with imports where the names are the keys and their modules are the values
        """

        imports = []
        aliases = []
        for node in ast.iter_child_nodes(ast.parse(import_string)):
            name = Imports._parse_node(node)
            if name is not None:
                imports += [(n[0], n[1]) for n in name]
                aliases += [(n[2], n[0]) for n in name if n[2] is not None]

        return dict(imports), dict(aliases)

    @staticmethod
    def _parse_node(node, module=''):
        """
        extract imports from ast nodes recursively
        :param node: ast nodes or list of nodes
        :param module: module of the imported name, is handled automatically by recursive calls
        :return: tuple of imported name and its module name
        """
        if isinstance(node, Sequence):
            return [Imports._parse_node(n, module) for n in node]
        elif isinstance(node, ast.Import) or isinstance(node, ast.ImportFrom):
            return Imports._parse_node(node.names, getattr(node, 'module', ''))
        elif isinstance(node, ast.alias):
            return node.name, module, node.asname

    def dump(self):
        """
        convert imports back to source code
        :return: python source code of the imports
        """

        text = ''
        for key, value in self.imports.items():
            if value != '':
                text += f'from {value} '
            text += f'import {key}\n'
        return text

    def is_imported(self, name):
        """
        check if name is imported
        :param name: name to check
        :return: bool
        """
        return name in self.imports.keys() or name in self.aliases.keys()


class Config(Imports):
    """
    class for handling .py config files
    """

    def __init__(self, imports='', file=None, **kwargs):
        """
        initialize Config instance, all kwargs become attributes of the instance
        :param imports: source code string or instance of Imports
        :param file: filename of the config module
        :param kwargs: parameters of the config
        """
        self.__dict__.update(kwargs.copy())
        self.parameters = list(kwargs.keys())
        super(Config, self).__init__(imports)
        self.file = file

    def __repr__(self):
        return self.dump()

    def __setattr__(self, key, value):
        """
        append names of new attrs to parameters
        """
        if key not in ('parameters', 'file', 'imports', 'aliases') and key not in self.parameters:
            self.parameters.append(key)
        return super(Config, self).__setattr__(key, value)

    @classmethod
    def from_module(cls, module):
        """
        initialize instance from python module
        :param module: python module
        :return: Config instance
        """

        imports = super(Config, cls).from_module(module)
        varnames = [var for var in dir(module) if not var.startswith('__') and not imports.is_imported(var)]
        variables = {varname: getattr(module, varname) for varname in varnames}

        return cls(imports=imports, file=module.__file__, **variables)

    def _parse_dict(self, dictionary):
        """
        parse a dict to valid python code
        :param dictionary: dict object
        :return: parsed dict as string
        """

        # define conversion function for variable types
        conversion = {str: lambda x: f'\'{x}\'',
                      type: lambda x: self._parse_type(x),
                      FunctionType: lambda x: self._parse_callable_in_dict(x),
                      BuiltinFunctionType: lambda x: self._parse_callable_in_dict(x),
                      dict: lambda x: self._parse_dict(x)}
        default = lambda x: x

        # convert variables in the dictionary
        output = {key: conversion.get(type(value), default)(value) for key, value in dictionary.items()}

        # construct string representation of the dictionary
        if not output:
            return str(output)

        text = '{'

        for key, value in output.items():
            text += f'\'{key}\': {value},\n'

        return text[:-2] + '}'

    def _parse_callable_in_dict(self, func):
        """
        parse callable in a dict to valid python code
        :param func: callable
        :return: parsed callable as string
        """

        LAMBDA = lambda: None
        if func.__name__ == LAMBDA.__name__:
            code = inspect.getsource(func).rstrip(' \n,')
            return code[code.find('lambda'):]
        elif isinstance(func, FunctionType) or isinstance(func, BuiltinFunctionType):
            return func.__name__
        elif isinstance(func, type):
            return self._parse_type(func)
        else:
            raise ValueError(f'{type(func)} objects cannot be handled by the conversion function')

    def _parse_type(self, obj):
        """
        parse class instance to valid python code
        :param obj: class instance
        :return: parsed instance as string
        """

        if self.is_imported(obj.__name__):
            return obj.__name__
        else:
            return str(obj).replace('class', '').lstrip('< \'').rstrip('> \'')

    def var_to_source_code(self, varname):
        """
        convert a variable from the config to valid python code
        :param varname: name of the variable
        :return: parsed variable as string
        """

        conversion = {str: lambda varname, var: f'{varname} = \'{var}\'',
                      dict: lambda varname, var: f'{varname} = {self._parse_dict(var)}',
                      type: lambda varname, var: f'{varname} = {self._parse_type(var)}',
                      FunctionType: lambda varname, var: f'{inspect.getsource(var)}',
                      BuiltinFunctionType: lambda varname, var: f'{inspect.getsource(var)}'}
        default = lambda varname, var: f'{varname} = {var}'

        var = getattr(self, varname)

        return f'\n{conversion.get(type(var), default)(varname, var)}\n'

    def dump(self):
        """
        dump the variables of a config module to a string
        :return: parsed config as a string
        """

        text = super(Config, self).dump() + '\n'

        for parameter in self.parameters:
            text += self.var_to_source_code(parameter)

        return text

    def save(self, filename, copy_file=True):
        """
        save the config module to a file
        :param filename: name under which the config will be saved
        :param copy_file: if True, copy the original file of the module, otherwise dump module's variables to text
        :return: None
        """

        if copy_file and self.file != '':
            shutil.copy(self.file, filename)
        else:
            with open(filename, 'w') as file:
                file.write(self.dump())
                self.file = filename if self.file == '' else self.file